#include "lua.h"
#include "lauxlib.h"

//#include "def.h"
#include "errfunc.h"

#include <SEGA_GFS.H>


// keep this info global
static GfsDirTbl gfsDirTbl;
static Sint32    gfsDirN;// number of files in current directory


Sint32 my_GFS_Load_malloc(Sint32 fid, Sint32 off, void **buf)
{
    GfsHn gfs;                 /* t@Cnh1 */
    Sint32 stat, sctsize, nsct, lastsize, bsize;

    /* t@CJ */
    gfs = GFS_Open(fid);
    if(gfs == NULL)
        return -1;

    GFS_GetFileSize(gfs, &sctsize, &nsct, &lastsize);
	bsize = sctsize * (nsct-1) + lastsize;
	*buf = (void *)my_malloc(bsize);
    if(*buf == NULL)
        return -2;

    if(GFS_Seek(gfs, off, GFS_SEEK_SET) < 0)
        return -3;
        
    /* t@Cǂݍ݊Jn */
    GFS_NwFread(gfs, nsct, *buf, bsize);
    for (;;) {
        stat = GFS_NwExecOne(gfs);
        if(stat == GFS_SVR_ERROR)
            return -4;
        else if(stat == GFS_SVR_COMPLETED)
            break;
//        user();
        slSynch();
    }
    /* t@C */
    GFS_Close(gfs);
    return (sctsize * (nsct-1) + lastsize);
}



//	Inputs
//		int DIR_MAX, int OPEN_MAX, GfsDirType (GFS_DIR_ID, GFS_DIR_NAME)
//	Outputs
//		on success	:	gfsDirName, gfsLibWork, gfsDirN
//		on error	:	none
//	Notes
//		keep returned tables stored, as long as GFS is in use (global variable)

static int lua_GFS_Init(lua_State *L)
{
	if(lua_gettop(L)==3 /*&& lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3)*/ ) {
	    // init FLD and GFS
	//    FLD_INIT_DDS();
	//    *((Uint32 *)GFS_DDS_ADDR) = 0x20202020;
	    GFS_DIRTBL_NDIR(&gfsDirTbl)    = lua_tointeger(L, 1);//DIR_MAX
	    GFS_DIRTBL_TYPE(&gfsDirTbl)    = lua_tointeger(L, 3);//GFS_DIR_ID, GFS_DIR_NAME;
		switch(lua_tointeger(L, 3)) {
			case GFS_DIR_ID:// = 0
			    GFS_DIRTBL_DIRID(&gfsDirTbl) = (GfsDirId *)lua_newuserdata(L, GFS_DIRTBL_NDIR(&gfsDirTbl) * sizeof(GfsDirId));
			    if(GFS_DIRTBL_DIRID(&gfsDirTbl) == NULL)
			    	return 0;
				break;
			case GFS_DIR_NAME:// = 1
			    GFS_DIRTBL_DIRNAME(&gfsDirTbl) = (GfsDirName *)lua_newuserdata(L, GFS_DIRTBL_NDIR(&gfsDirTbl) * sizeof(GfsDirName));
			    if(GFS_DIRTBL_DIRNAME(&gfsDirTbl) == NULL)
			    	return 0;
				break;
		}
		Uint32 *gfsLibWork = (Uint32 *)lua_newuserdata(L, GFS_WORK_SIZE(lua_tointeger(L, 2)));
		gfsDirN = GFS_Init(lua_tointeger(L, 2), gfsLibWork, &gfsDirTbl);
	    if(gfsDirN > 0) {
			lua_pushinteger(L, gfsDirN);
		    GFS_SetErrFunc(errGfsFunc, NULL);
			return 3;
	    } else lua_pop(L, 1);// free: GfsDirName
	}
	return 0;// fail
}





// taken from smppcm4.c
static GfsHn fileOpen(char *fname)
{
    Sint32 fid;
//    int i, n = -1;
//    for(i=0; i < 14; i++) {
//    	if(fname[i] == '\0')
//    		n = i;
//	}
//    /* t@Ct@Cʎq߂ */
//	slPrintHex(n, slLocate(5, 20));
//	slPrint(fname, slLocate(5, 21));
    fid = GFS_NameToId((Sint8 *)fname);
    return GFS_Open(fid);
//	slPrintHex(fid, slLocate(5, 22));
////	slPrintHex(GFS_Open(fid), slLocate(5, 23));// GFS_open fucks it up
//	return NULL;
}




//	Inputs
//		char *filename
//	Outputs
//		on success	:	GfsHn
//		on error	:	none

static int lua_fileOpen(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_isstring(L, 1)) {
		GfsHn gfs = fileOpen((char *)lua_tostring(L, 1)); // yabause crashes right here
		if(gfs != NULL) {
			lua_pushlightuserdata(L, gfs);
//	slPrint("fileOpen: OK", slLocate(5, 11));
//	while(1) slSynch();
			return 1;
		}
	}
//	slPrint("fileOpen: ERROR", slLocate(5, 11));
//	while(1) slSynch();
	return 0;
}

//	Inputs
//		Sint32 fid, Sint32 ofs, void *buf, Sint32 bsize
//	Outputs
//		on success:	number of bytes loaded
//		on error:	negative error code / no return value

static int lua_GFS_Load(lua_State *L)
{
	if(lua_gettop(L)==4 /*&& lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 4)*/ ) {
		void *b = lua_isuserdata(L, 3) ? (void *)lua_touserdata(L, 3) : (void *)lua_tointeger(L, 3);
		if(b != NULL) {
			lua_pushinteger(L, GFS_Load((Sint32)lua_tointeger(L, 1), (Sint32)lua_tointeger(L, 2), b, (Sint32)lua_tointeger(L, 4)));
			return 1;
		}
	}
	return 0;
}



//	Inputs
//		char *fname, Sint32 ofs, void *buf, Sint32 bsize
//	Outputs
//		on success:	number of bytes loaded
//		on error:	negative error code / no return value

static int lua_GFS_Load_name(lua_State *L)
{
	if(lua_gettop(L)==4 && lua_isstring(L, 1) /*&& lua_isnumber(L, 2) && lua_isnumber(L, 4)*/ ) {
		void *b = lua_isuserdata(L, 3) ? (void *)lua_touserdata(L, 3) : (void *)lua_tointeger(L, 3);
		if(b != NULL) {
			lua_pushinteger(L, GFS_Load(GFS_NameToId((Sint8 *)lua_tostring(L, 1)), (Sint32)lua_tointeger(L, 2), b, (Sint32)lua_tointeger(L, 4)));
			return 1;
		}
	}
	return 0;
}



Sint32 lua_GFS_Load_malloc(lua_State *L, Sint32 fid, Sint32 off)
{
    GfsHn gfs;                 /* t@Cnh1 */
    Sint32 stat, sctsize, nsct, lastsize, bsize;
    void *buf;
//    int ret, i = 0;
//    Sint32 ndata;
//    GfsHn gfs_now;

    /* t@CJ */
    gfs = GFS_Open(fid);
    if(gfs == NULL)
        return -1;

    GFS_GetFileSize(gfs, &sctsize, &nsct, &lastsize);
	bsize = sctsize * (nsct-1) + lastsize;
	buf = (void *)lua_newuserdata(L, bsize);
    if(buf == NULL)
        return -2;
    lua_pushinteger(L, bsize);

    if(GFS_Seek(gfs, off, GFS_SEEK_SET) < 0)
        return -3;
        
    /* t@Cǂݍ݊Jn */
    GFS_NwFread(gfs, nsct, buf, bsize);
    for (;;) {
        stat = GFS_NwExecOne(gfs);
//        stat = GFS_NwExecServer(&gfs_now);
        if(stat == GFS_SVR_ERROR)
            return -4;
        else if(stat == GFS_SVR_COMPLETED)
            break;
//        user();
		// stat = 1, file pos stays at 0
		// ? GFS_NwCdFread() instead ?
		// SCU DMA 
//		slPrint("GFS_Load_malloc() Status:", slLocate(5, 20));
//		slPrintHex(stat, slLocate(5, 21));
//		slPrintHex(GFS_Tell(gfs), slLocate(5, 22));
//		slPrintHex(gfs, slLocate(5, 23));
//		slPrintHex(gfs_now, slLocate(25, 23));// indeed, this one id different
//		slPrintHex(ret, slLocate(5, 24));
//		slPrintHex(i++, slLocate(5, 25));
//		slPrintHex(GFS_NwIsComplete(gfs), slLocate(5, 26));
//		GFS_NwGetStat(gfs, &stat, &ndata);
//		slPrintHex(stat, slLocate(5, 27));
//		slPrintHex(ndata, slLocate(25, 27));
        slSynch();
    }
    /* t@C */
// didn't help	GFS_NwStop(gfs);// trying to fix PCM not continueing after loading a file
    GFS_Close(gfs);
    return (sctsize * (nsct-1) + lastsize);
}



//	Inputs
//		Sint32 fid, Sint32 ofs
//	Outputs
//		on success:	void *buf, Sint32 bsize, number of bytes loaded
//		on error:	negative error code as bytenum / no return value

static int lua_GFS_Load_buffer(lua_State *L)
{
	if(lua_gettop(L)==2 /*&& lua_isnumber(L, 1) && lua_isnumber(L, 2)*/ ) {
		Sint32 ret = lua_GFS_Load_malloc(L, (Sint32)lua_tointeger(L, 1), (Sint32)lua_tointeger(L, 2));
		if(ret != -1 && ret != -2) {
			lua_pushinteger(L, ret);
			return 3;
		}
	}
	return 0;
}



//	Inputs
//		char *fname, Sint32 ofs
//	Outputs
//		on success:	void *buf, Sint32 bsize, number of bytes loaded
//		on error:	negative error code as bytenum / no return value

static int lua_GFS_Load_name_buffer(lua_State *L)
{
	if(lua_gettop(L)==2 && lua_isstring(L, 1) /*&& lua_isnumber(L, 2)*/ ) {
		Sint32 ret = lua_GFS_Load_malloc(L, GFS_NameToId((Sint8 *)lua_tostring(L, 1)), (Sint32)lua_tointeger(L, 2));
		if(ret != -1 && ret != -2) {
			lua_pushinteger(L, ret);
			return 3;
		}
	}
	return 0;
}

//	Inputs
//		Sint32 fid
//	Outputs
//		on success:	file size
//		on error:	none

static int lua_GFS_DIR_SIZE(lua_State *L)
{
	if(lua_gettop(L)==1 /*&& lua_isnumber(L, 1)*/ ) {
		Sint32 fid = lua_tointeger(L, 1);
		if(fid<GFS_DIRTBL_NDIR(&gfsDirTbl) && fid<gfsDirN) {
			switch(GFS_DIRTBL_TYPE(&gfsDirTbl)) {
				case GFS_DIR_ID:
					lua_pushinteger(L, GFS_DIR_SIZE( GFS_DIRTBL_DIRID(&gfsDirTbl) + fid ));
					return 1;
				case GFS_DIR_NAME:
					lua_pushinteger(L, GFS_DIR_SIZE( GFS_DIRTBL_DIRNAME(&gfsDirTbl) + fid ));
					return 1;
			}		
		}
	}
	return 0;
}






//	Inputs
//		GfsHn gfs
//	Outputs
//		on success:	Sint32 amode, Sint32 ndata
//		on error:	none

static int lua_GFS_NwGetStat(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1)) {
		Sint32 amode, ndata;
		GFS_NwGetStat((GfsHn)lua_touserdata(L, 1), &amode, &ndata);
		lua_pushinteger(L, amode);
		lua_pushinteger(L, ndata);
		return 2;
	}
	return 0;
}



//	Inputs
//		none
//	Outputs
//		on success:	Sint32 return value, GfsHn now_gfs
//		on error:	none

static int lua_GFS_NwExecServer(lua_State *L)
{
	GfsHn now_gfs;
	lua_pushinteger(L, GFS_NwExecServer(&now_gfs));
	lua_pushlightuserdata(L, now_gfs);
	return 2;
}







void lua_GFS_register(lua_State *L)
{
	if(L != NULL) {
		lua_register(L, "GFS_Init" 				, lua_GFS_Init);
		lua_register(L, "fileOpen" 				, lua_fileOpen);
		lua_register(L, "GFS_Load"				, lua_GFS_Load);
		lua_register(L, "GFS_Load_name"			, lua_GFS_Load_name);
		lua_register(L, "GFS_Load_buffer"		, lua_GFS_Load_buffer);
		lua_register(L, "GFS_Load_name_buffer"	, lua_GFS_Load_name_buffer);
		lua_register(L, "GFS_DIR_SIZE"			, lua_GFS_DIR_SIZE);
		lua_register(L, "GFS_NwGetStat"			, lua_GFS_NwGetStat);
		lua_register(L, "GFS_NwExecServer"		, lua_GFS_NwExecServer);
	}
}
