#include "lua.h"
#include "lauxlib.h"

//#include "def.h"

// overrides gcc command line define
//#ifndef USE_ADPCM
//#	define USE_ADPCM
//#endif


#include "errfunc.h"
#include "alloc.h"

#include <SEGA_STM.H>
#include <SEGA_PCM.H>

//#include "vblank.h"

//	Inputs
//		none
//	Outputs
//		success:	Boolean result of PCM_Init()
//		error:		none

static int lua_PCM_Init(lua_State *L)
{
	lua_pushboolean(L, PCM_Init());
#ifdef USE_ADPCM
	PCM_DeclareUseAdpcm();
#endif
	PCM_SetErrFunc(errPcmFunc, NULL);
// is now added manually in Lua	
//	if(vblk_func_add(PCM_VblIn) >= 0) {
////		slPrint("success", slLocate(3, 19));
//		return 1;
//	}
//	else lua_pop(L, 1);
////	slPrint("fail", slLocate(3, 19));
//	return 0;
	return 1;
}



//	Inputs
//		none
//	Outputs
//		none

static int lua_PCM_Finish(lua_State *L)
{
	PCM_Finish();
	return 0;
}



//	Inputs
//		unsigned int pcm_size, Sint8 *pcm_addr, unsigned int ring_size [, Sint8 *ring_addr]
//			#define	PCM_ADDR_1	((void*)0x25a20000)
//			#define	PCM_ADDR_2	((void*)0x25a40000)
//	Outputs
//		number of items put on stack: 2/1 on success, 0 on error

static int PcmCreatePara_new(lua_State *L, PcmCreatePara *para, char allow_4_parameters)
{
	if(lua_gettop(L)>=3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3)) {
		// need buffer: struct PcmWork
		PCM_PARA_WORK(para) = (struct PcmWork *)lua_newuserdata(L, sizeof(PcmWork));
		if(PCM_PARA_WORK(para) != NULL) {
			PCM_PARA_PCM_SIZE(para)  = (Sint32)lua_tointeger(L, 1);
			PCM_PARA_PCM_ADDR(para)  = (Sint8 *)lua_tointeger(L, 2);
			PCM_PARA_RING_SIZE(para) = (Sint32)lua_tointeger(L, 3);
			// need buffer: ring buffer
			if(allow_4_parameters==1 &&lua_gettop(L)>=4 && lua_isuserdata(L, 4)) {
				// use given ring buffer
				PCM_PARA_RING_ADDR(para) = (Sint8 *)lua_touserdata(L, 4);
//slPrintHex((unsigned int)lua_tointeger(L, 3), slLocate(35, 5));
//slPrintHex((unsigned int)lua_touserdata(L, 4), slLocate(35, 6));
				return 1;//success
			} else {
				// allocate new ring buffer
				PCM_PARA_RING_ADDR(para) = (Sint8 *)lua_newuserdata(L, PCM_PARA_RING_SIZE(para));
				if(PCM_PARA_RING_ADDR(para) != NULL) {
//					slPrint("success2", slLocate(3, 19));
					return 2;//success
				} else lua_pop(L, 1);// free: struct PcmWork
			}
		}
	}
//	slPrint("fail2", slLocate(3, 19));
	return 0;//fail
}



//	Inputs
//		unsigned int pcm_size, Sint8 *pcm_addr, unsigned int ring_size, GfsHn
//	Outputs
//		on success:	struct PcmWork, Sint *ring_buffer, PcmHn
//		on error:	none

static int lua_PCM_CreateGfsHandle(lua_State *L)
{
	if(lua_gettop(L)==4 && lua_islightuserdata(L, 4)) {
		PcmCreatePara	para;// I assume that structure is copied into work structure and thus is not used anymore afterwards		
		if(PcmCreatePara_new(L, &para, 0)) {
			PcmHn pcm = PCM_CreateGfsHandle(&para, (GfsHn)lua_touserdata(L, 4));
			if(pcm != NULL) {
				lua_pushlightuserdata(L, pcm);
//				slPrint("success", slLocate(3, 20));
				return 3;
			} else lua_pop(L, 2);// free: struct PcmWork, ring_buffer
		}
	}
//	slPrint("fail", slLocate(3, 20));
	return 0;
}



//	Inputs
//		unsigned int pcm_size, Sint8 *pcm_addr, unsigned int ring_size, StmHn
//	Outputs
//		on success:	struct PcmWork, Sint *ring_buffer, PcmHn
//		on error:	none

static int lua_PCM_CreateStmHandle(lua_State *L)
{
	if(lua_gettop(L)==4 && lua_islightuserdata(L, 4)) {
		PcmCreatePara	para;// I assume that structure is copied into work structure and thus is not used anymore afterwards
		
		if(PcmCreatePara_new(L, &para, 0)) {
			PcmHn pcm = PCM_CreateStmHandle(&para, (StmHn)lua_touserdata(L, 4));
			if(pcm != NULL) {
				lua_pushlightuserdata(L, pcm);
				return 3;
			} else lua_pop(L, 2);// free: struct PcmWork, ring_buffer
		}
	}
	return 0;
}



//	Inputs
//		unsigned int pcm_size, Sint8 *pcm_addr, unsigned int ring_size [, Sint8 *ring_addr]
//	Outputs
//		on success:	struct PcmWork[, Sint *ring_buffer], PcmHn
//		on error:	none
//	Notes
//		Specify file size as ring_size, afterwards place file into ring_buffer. If file already loaded, specify it's address as ring_addr.

static int lua_PCM_CreateMemHandle(lua_State *L)
{
	PcmCreatePara	para;// I assume that structure is copied into work structure and thus is not used anymore afterwards
	int ret = PcmCreatePara_new(L, &para, 1);
	if(ret > 0) {
		PcmHn pcm = PCM_CreateMemHandle(&para);
		if(pcm != NULL) {
			lua_pushlightuserdata(L, pcm);
			PCM_NotifyWriteSize(pcm, (Sint32)lua_tointeger(L, 1));
			return ret+1;
		} else lua_pop(L, ret);// free: struct PcmWork[, ring_buffer]
	}
	return 0;
}



//	Inputs
//		PcmHn
//	Outputs
//		none

static int lua_PCM_DestroyGfsHandle(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1))
		PCM_DestroyGfsHandle((PcmHn)lua_touserdata(L, 1));
	return 0;
}



//	Inputs
//		PcmHn
//	Outputs
//		none

static int lua_PCM_DestroyStmHandle(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1))
		PCM_DestroyStmHandle((PcmHn)lua_touserdata(L, 1));
	return 0;
}



//	Inputs
//		PcmHn
//	Outputs
//		none

static int lua_PCM_DestroyMemHandle(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1))
		PCM_DestroyMemHandle((PcmHn)lua_touserdata(L, 1));
	return 0;
}



//	Inputs
//		PcmHn pcm, Sint32 cnt_loop
//	Outputs
//		none

static int lua_PCM_SetLoop(lua_State *L)
{
	if(lua_gettop(L)==2 && lua_islightuserdata(L, 1) /*&& lua_isnumber(L, 2)*/ )
		PCM_SetLoop((PcmHn)lua_touserdata(L, 1), (Sint32)lua_tointeger(L, 2));
	return 0;
}



//	Inputs
//		PcmHn
//	Outputs
//		none

static int lua_PCM_Start(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1))
		PCM_Start((PcmHn)lua_touserdata(L, 1));
	return 0;
}



//	Inputs
//		PcmHn
//	Outputs
//		none

static int lua_PCM_Stop(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1))
		PCM_Stop((PcmHn)lua_touserdata(L, 1));
	return 0;
}



//	Inputs
//		PcmHn, Boolean pause true/false (= on/off)
//	Outputs
//		Boolean success
//	Notes
//		temporarly allocates a pause work buffer, uses my_malloc() and my_free()

static int lua_PCM_Pause(lua_State *L)
{
	int ret = 0;
	
	if(lua_gettop(L)==2 && lua_isuserdata(L, 1) && lua_isboolean(L, 2)) {
		// PAUSE On or OFF?
		if(lua_toboolean(L, 2) == 0) {	// PCM_PAUSE_OFF
			PCM_Pause((PcmHn)lua_touserdata(L, 1), PCM_PAUSE_OFF);
			ret = 1;
		} else {						// PCM_PAUSE_ON_AT_ONCE
			// Pause work size: min 4096, optimal when same size as PCM buffer
			// try allocating as much as possible, cut size by half, when failed
			int   pwork_size;
			void *pwork_addr;
			for(pwork_size = 4*4096; pwork_size >= 4096; pwork_size >>= 1) {
				pwork_addr = my_malloc(pwork_size);
				if(pwork_addr != NULL) {
					PCM_SetPauseWork(pwork_addr, pwork_size);
					PCM_Pause((PcmHn)lua_touserdata(L, 1), PCM_PAUSE_ON_AT_ONCE);
					my_free(pwork_addr);
					ret = 1;
					break;
				}
			}
		}
	}
	// output
	lua_pushboolean(L, ret);
	return 1;
}



//	Inputs
//		[PcmHn | NULL/0/nil]
//	Outputs
//		none

static int lua_PCM_Task(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1))
		PCM_Task((PcmHn)lua_touserdata(L, 1));
	else /*if(lua_gettop(L)==0 ||
			(lua_gettop(L)==1 &&
				(lua_isnil(L, 1) ||
					(lua_isnumber(L, 1) && lua_tointeger(L, 1)==0)
				)
			)
		)*/
		PCM_Task(NULL);

	return 0;
}



//	Inputs
//		PcmHn
//	Outputs
//		on success	: Integer status
//		on error	: none

static int lua_PCM_GetPlayStatus(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1)) {
		lua_pushinteger(L, PCM_GetPlayStatus((PcmHn)lua_touserdata(L, 1)));
		return 1;
	} else
		return 0;
}



//	Inputs
//		PcmHn pcm, Sint32 stream_no
//	Outputs
//		none

static int lua_PCM_SetPcmStreamNo(lua_State *L)
{
	if(lua_gettop(L)==2 && lua_islightuserdata(L, 1) /*&& lua_isnumber(L, 2)*/ ) {
//		slPrint("Set stream number", slLocate(5, 9+lua_tointeger(L, 2)));
		PCM_SetPcmStreamNo((PcmHn)lua_touserdata(L, 1), (Sint32)lua_tointeger(L, 2));
	}
	return 0;
}



//	Inputs
//		PcmHn pcm
//	Outputs
//		on success: write_adr, free_size, free_total
//		on fail:    none

static int lua_PCM_GetWriteBuf(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1)) {
		Sint32 free_size, free_total;		
		lua_pushlightuserdata(L, PCM_GetWriteBuf((PcmHn)lua_touserdata(L, 1), &free_size, &free_total));
		lua_pushinteger(L, free_size);
		lua_pushinteger(L, free_total);
		return 3;
	}
	return 0;
}



//	Inputs
//		PcmHn pcm, Sint32 write_size
//	Outputs
//		none

static int lua_PCM_NotifyWriteSize(lua_State *L)
{
	if(lua_gettop(L)==2 && lua_islightuserdata(L, 1) /*&& lua_isnumber(L, 2)*/ ) {		
		PCM_NotifyWriteSize((PcmHn)lua_touserdata(L, 1), (Sint32)lua_tointeger(L, 2));
	}
	return 0;
}




//	Inputs
//		PcmHn
//	Outputs
//		on success	: channel, sampling_bit, sampling_rate, sample_file
//		on error	: none

static int lua_PCM_GetInfo(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_isuserdata(L, 1)) {//lua_islightuserdata(L, 1)) {
// false		PcmInfo *info = &(((PcmStatus*)lua_touserdata(L, 1))->info);
// 		PcmInfo *info = &(((*(PcmWork **)lua_touserdata(L, 1)))->status.info);// returns zeros
		PcmInfo *info = &(((PcmWork*)lua_touserdata(L, 1))->status.info);// returns zeros
		lua_pushinteger(L, info->channel);
		lua_pushinteger(L, info->sampling_bit);
		lua_pushinteger(L, info->sampling_rate);
		lua_pushinteger(L, info->sample_file);
		return 4;
	} else
		return 0;
}




//	Inputs
//		PcmHn pcm, Sint32 volume
//	Outputs
//		none

static int lua_PCM_SetVolume(lua_State *L)
{
	if(lua_gettop(L)==2 && lua_islightuserdata(L, 1) /*&& lua_isnumber(L, 2)*/ ) {		
		PCM_SetVolume((PcmHn)lua_touserdata(L, 1), (Sint32)lua_tointeger(L, 2));
	}
	return 0;
}




//	Inputs
//		PcmHn pcm
//	Outputs
//		none

static int lua_PCM_ChangePcmPara(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1)) {		
		PCM_ChangePcmPara((PcmHn)lua_touserdata(L, 1));
	}
	return 0;
}



void lua_PCM_register(lua_State *L)
{
	if(L != NULL) {
		lua_register(L, "PCM_Init" 				, lua_PCM_Init);
		lua_register(L, "PCM_Finish" 			, lua_PCM_Finish);
		lua_register(L, "PCM_SetLoop" 			, lua_PCM_SetLoop);
		lua_register(L, "PCM_Start" 			, lua_PCM_Start);
		lua_register(L, "PCM_Stop" 				, lua_PCM_Stop);
		lua_register(L, "PCM_Pause" 			, lua_PCM_Pause);
		lua_register(L, "PCM_Task" 				, lua_PCM_Task);
		lua_register(L, "PCM_GetPlayStatus" 	, lua_PCM_GetPlayStatus);
		lua_register(L, "PCM_CreateGfsHandle" 	, lua_PCM_CreateGfsHandle);
		lua_register(L, "PCM_CreateStmHandle" 	, lua_PCM_CreateStmHandle);
		lua_register(L, "PCM_CreateMemHandle" 	, lua_PCM_CreateMemHandle);
		lua_register(L, "PCM_DestroyGfsHandle" 	, lua_PCM_DestroyGfsHandle);
		lua_register(L, "PCM_DestroyStmHandle" 	, lua_PCM_DestroyStmHandle);
		lua_register(L, "PCM_DestroyMemHandle" 	, lua_PCM_DestroyMemHandle);
		lua_register(L, "PCM_SetPcmStreamNo" 	, lua_PCM_SetPcmStreamNo);
		lua_register(L, "PCM_GetWriteBuf" 		, lua_PCM_GetWriteBuf);
		lua_register(L, "PCM_NotifyWriteSize" 	, lua_PCM_NotifyWriteSize);
		lua_register(L, "PCM_GetInfo" 			, lua_PCM_GetInfo);
		lua_register(L, "PCM_SetVolume" 		, lua_PCM_SetVolume);
		lua_register(L, "PCM_ChangePcmPara" 	, lua_PCM_ChangePcmPara);
	}
}
