#include "lua.h"
#include "lauxlib.h"

//#include "def.h"

#include <SGL.H>
#include <SEGA_SYS.H>// for clock change

#include "api_def_implementation.h"// just for definition of fromFIXED()

extern void getVECTORfromTable(lua_State *L, int idx, FIXED *vct);
extern void setVECTORtoTable(lua_State *L, FIXED *vct);
extern void setfield_Integer(lua_State *L, unsigned int v, const char *k);
extern void setfield_Number(lua_State *L, float v, const char *k);

/****************************/
/*  Text display library	*/
/****************************/


//	Inputs
//		string, xpos, ypos
//	Outputs
//		on success:	void *buffer
//		on error:	none

static int lua_slPrint(lua_State *L)
{
	if(lua_gettop(L)==3 && lua_isstring(L, 1) /*&& lua_isnumber(L, 2) && lua_isnumber(L, 3)*/ ) {
		slPrint((char *)lua_tostring(L, 1), slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)));
		lua_pushboolean(L, 1);// success
	} else lua_pushboolean(L, 0);
	return 1;
}



//	Inputs
//		integer, xpos, ypos
//	Outputs
//		on success:	void *buffer
//		on error:	none

static int lua_slPrintHex(lua_State *L)
{
//	if(lua_gettop(L)==3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3)) {
//		slPrintHex(lua_tointeger(L, 1), slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)));
// allow display of memory locations
	if(lua_gettop(L)==3 /*&& lua_isnumber(L, 2) && lua_isnumber(L, 3)*/ ) {
		int val = lua_tointeger(L, 1);
		if (lua_isuserdata(L, 1))
			val = (int)lua_touserdata(L, 1);
		slPrintHex(val, slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)));
		lua_pushboolean(L, 1);// success
	} else lua_pushboolean(L, 0);
	return 1;
}



//	Inputs
//		fixed, xpos, ypos
//	Outputs
//		on success:	void *buffer
//		on error:	none

static int lua_slPrintFX(lua_State *L)
{
	if(lua_gettop(L)==3 /*&& lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3)*/ ) {
		slPrintFX(toFIXED((float)lua_tonumber(L, 1)), slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)));
		lua_pushboolean(L, 1);// success
	} else lua_pushboolean(L, 0);
	return 1;
}


//	Inputs
//		integer, xpos, ypos
//	Outputs
//		on success:	void *buffer
//		on error:	none

static int lua_slDispHex(lua_State *L)
{
//	if(lua_gettop(L)==3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3)) {
//		slPrintHex(lua_tointeger(L, 1), slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)));
// allow display of memory locations
	if(lua_gettop(L)==3 /*&& lua_isnumber(L, 2) && lua_isnumber(L, 3)*/ ) {
		int val = lua_tointeger(L, 1);
		if (lua_isuserdata(L, 1))
			val = (int)lua_touserdata(L, 1);
		slDispHex(val, slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)));
		lua_pushboolean(L, 1);// success
	} else lua_pushboolean(L, 0);
	return 1;
}

//	Inputs
//		MATRIX pointer, xpos, ypos
//	Outputs
//		on success:	void *buffer
//		on error:	none

static int lua_slPrintMatrix(lua_State *L)
{
//	if(lua_gettop(L)==3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3)) {
//		slPrintHex(lua_tointeger(L, 1), slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)));
// allow display of memory locations
	if(lua_gettop(L)==3 /*&& lua_isnumber(L, 2) && lua_isnumber(L, 3)*/ ) {
		void *val = lua_isuserdata(L, 1) ? (void *)lua_touserdata(L, 1) : (void *)lua_tointeger(L, 1);
		if(val != NULL) {
			slPrintMatrix(val, slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)));
			lua_pushboolean(L, 1);// success
		}
	} else lua_pushboolean(L, 0);
	return 1;
}



// Note:  3 inputs: this version takes binaries compiled with GCC
// Note: >3 inputs: this version takes commands specified in Lua
//	Inputs
//		Uint8 array of commands, ... x, y
//	Outputs
//		void   *

static int lua_slExtendScroll(lua_State *L)
{
	if(lua_gettop(L)==3 /*&& lua_isnumber(L, 2) && lua_isnumber(L, 3)*/ ) {
		Uint8 *p = lua_isuserdata(L, 1) ? (Uint8 *)lua_touserdata(L, 1) : (Uint8 *)lua_tointeger(L, 1);
		if(p != NULL)
			slExtendScroll( p, slLocate(lua_tointeger(L, 2), lua_tointeger(L, 3)) );
	} else if(lua_gettop(L)>3 && lua_isnumber(L, 1) && lua_isnumber(L, 2)) {
		int i = 0, n = lua_gettop(L)-2;
		Uint8 cmd[n];
		for(i=0;i<n;i++)
			cmd[i] = (Uint8)lua_tointeger(L, i+1);
		slExtendScroll( cmd, slLocate(lua_tointeger(L, n+1), lua_tointeger(L, n+2)));
	}
	return 0;
}


/********************/
/*  Scroll library	*/
/********************/

// ToDo: add table support
//	Inputs
//		Uint8 map[16]
//	Outputs
//		none

//static int lua_sl16MapRA(lua_State *L)
//{
//	if(lua_gettop(L)==16) {
//		Uint8 map[16];
//		int i;
//		for(i=0;i<16;i++)
//			map[i] = (Uint8)lua_tointeger(L, i+1);
//		sl16MapRA( map );
//	}
//	return 0;
//}

static int lua_sl16MapRA(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_istable(L, 1)) {
		Uint8 map[16];
		int i;
		for(i=0;i<16;i++) {
			lua_pushinteger(L, i+1);
			lua_gettable(L, 1);
			map[i] = (Uint8)lua_tointeger(L, -1);
			lua_pop(L, 1);
		}
		sl16MapRA( map );
	}
	return 0;
}

// ToDo: add table support
//	Inputs
//		Uint8 map[16]
//	Outputs
//		none

//static int lua_sl16MapRB(lua_State *L)
//{
//	if(lua_gettop(L)==16) {
//		Uint8 map[16];
//		int i;
//		for(i=0;i<16;i++)
//			map[i] = (Uint8)lua_tointeger(L, i+1);
//		sl16MapRB( map );
//	}
//	return 0;
//}
static int lua_sl16MapRB(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_istable(L, 1)) {
		Uint8 map[16];
		int i;
		for(i=0;i<16;i++) {
			lua_pushinteger(L, i+1);
			lua_gettable(L, 1);
			map[i] = (Uint8)lua_tointeger(L, -1);
			lua_pop(L, 1);
		}
		sl16MapRB( map );
	}
	return 0;
}

//	Inputs
//		(void *)data, (Uint32)size, channel number, bit depth, pitch
//	Outputs
//		PCM *pdat, return value of slPCMOn

static int lua_slPCMOn(lua_State *L)
{
	if(lua_gettop(L)==5 /*&& lua_isnumber(L, 2) && lua_isnumber(L, 3) && lua_isnumber(L, 4) && lua_isnumber(L, 5)*/ ) {
		void *data = lua_isuserdata(L, 1) ? (void *)lua_touserdata(L, 1) : (void *)lua_tointeger(L, 1);
		if(data != NULL) {
			// create the PCM structure
			PCM *pdat = (PCM *)lua_newuserdata(L, sizeof(PCM));
			pdat->mode 		= lua_tointeger(L, 3) == 1 ? _Mono : _Stereo ;		/* Mode */
			pdat->mode 		|= lua_tointeger(L, 4) == 8 ? _PCM8Bit : _PCM16Bit ;		/* Mode */
			pdat->channel 	= 0;	/* PCM Channel Number */
			pdat->level 	= 127;		/* 0 ~ 127 */
			pdat->pan 		= 0;		/* -128 ~ +127 */
			pdat->pitch 	= (Uint16)lua_tointeger(L, 5);
			pdat->eflevelR 	= 0;	/* Effect level for Right(mono) 0 ~ 7 */
			pdat->efselectR = 0;	/* Effect select for Right(mono) 0 ~ 15 */
			pdat->eflevelL 	= 0;	/* Effect level for Left 0 ~ 7 */
			pdat->efselectL = 0;	/* Effect select for Left 0 ~ 15 */
	
			lua_pushboolean(L, slPCMOn(	pdat,
										data,
										(Uint32)lua_tointeger(L, 2)) );
			return 2;
		}
	}
	return 0;
}


//	Inputs
//		PCM *pdat
//	Outputs
//		return value of slPCMOff

static int lua_slPCMOff(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_isuserdata(L, 1)) {
		lua_pushboolean(L, slPCMOff( (PCM *)lua_touserdata(L, 1)) );
		return 1;
	}
	return 0;
}


//	Inputs
//		PCM *pdat
//	Outputs
//		return value of slPCMStat

static int lua_slPCMStat(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_isuserdata(L, 1)) {
		lua_pushboolean(L, slPCMStat( (PCM *)lua_touserdata(L, 1)) );
		return 1;
	}
	return 0;
}

//	Inputs
//		Uint16 , TEXTURE * , Sint8
//	Outputs
//		none


static int lua_slInitSystem(lua_State *L)
{
	if(lua_gettop(L)==3 && lua_isnumber(L,1) && lua_isnumber(L,3)) {
		TEXTURE *t = lua_isuserdata(L,2) ? (TEXTURE *)lua_touserdata(L,2) : (TEXTURE *)lua_tointeger(L,2);
		slInitSystem((Uint16)lua_tointeger(L,1), t, (Sint8)lua_tointeger(L,3));
	}
	return 0;
}


/********************/
/*  Math library	*/
/********************/

//	Inputs
//		none
//	Outputs
//		X, Y, Z translation

//static int lua_slGetTranslate(lua_State *L)
//{
//	FIXED pbuf[XYZ];
//	slGetTranslate(pbuf);
//	lua_pushnumber(L, (float)pbuf[X] / 65536.0);
//	lua_pushnumber(L, (float)pbuf[Y] / 65536.0);
//	lua_pushnumber(L, (float)pbuf[Z] / 65536.0);
//	return 3;
//}
static int lua_slGetTranslate(lua_State *L)
{
	FIXED pbuf[XYZ];
	slGetTranslate(pbuf);
	setVECTORtoTable(L, pbuf);
	return 1;
}

//	Inputs
//		FIXED camera[XYZ], FIXED target[XYZ], ANGLE angz
//	Outputs
//		none

//static int lua_slLookAt(lua_State *L)
//{
//	if(lua_gettop(L)==7 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) && lua_isnumber(L, 4) && lua_isnumber(L, 5) && lua_isnumber(L, 6) && lua_isnumber(L, 7) ) {
//		FIXED camera[XYZ] = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		FIXED target[XYZ] = {toFIXED((float)lua_tonumber(L, 4)), toFIXED((float)lua_tonumber(L, 5)), toFIXED((float)lua_tonumber(L, 6))};
//		slLookAt(camera, target, (ANGLE)lua_tointeger(L, 7));
//	}
//	return 0;
//}

static int lua_slLookAt(lua_State *L)
{
	if(lua_gettop(L)==3 && lua_istable(L, 1) && lua_istable(L, 2) /*&& lua_isnumber(L, 3)*/ ) {
		FIXED camera[XYZ], target[XYZ];
		getVECTORfromTable(L, 1, camera);
		getVECTORfromTable(L, 2, target);
		slLookAt(camera, target, (ANGLE)lua_tointeger(L, 3));
	}
	return 0;
}

//	Inputs
//		VECTOR vct1,VECTOR vct2
//	Outputs
//		FIXED

//static int lua_slInnerProduct(lua_State *L)
//{
//	if(lua_gettop(L)==6 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) && lua_isnumber(L, 4) && lua_isnumber(L, 5) && lua_isnumber(L, 6) ) {
//		VECTOR vct1 = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		VECTOR vct2 = {toFIXED((float)lua_tonumber(L, 4)), toFIXED((float)lua_tonumber(L, 5)), toFIXED((float)lua_tonumber(L, 6))};
//		lua_pushnumber(L, (float)slInnerProduct(vct1, vct2) / 65536.0);
//		return 1;
//	}
//	return 0;
//}

static int lua_slInnerProduct(lua_State *L)
{
	if(lua_gettop(L)==2 && lua_istable(L, 1) && lua_istable(L, 2) ) {
		VECTOR vct1, vct2;
		getVECTORfromTable(L, 1, vct1);
		getVECTORfromTable(L, 2, vct2);
		lua_pushnumber(L, fromFIXED(slInnerProduct(vct1, vct2) ) );
		return 1;
	}
	return 0;
}

//	Inputs
//		none
//	Outputs
//		FIXED

static int lua_slRandom(lua_State *L)
{
	lua_pushnumber(L, fromFIXED(slRandom() ) );
	return 1;
}

//	Inputs
//		VECTOR p0, VECTOR p1, VECTOR p2, VECTOR p3, FIXED timing
//	Outputs
//		VECTOR ans

//static int lua_slBezier(lua_State *L)
//{
//	if(lua_gettop(L)==13 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) && lua_isnumber(L, 4) && lua_isnumber(L, 5) && lua_isnumber(L, 6) && lua_isnumber(L, 7) && lua_isnumber(L, 8) && lua_isnumber(L, 9) && lua_isnumber(L, 10) && lua_isnumber(L, 11) && lua_isnumber(L, 12) && lua_isnumber(L, 13) ) {
//		VECTOR p0 = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		VECTOR p1 = {toFIXED((float)lua_tonumber(L, 4)), toFIXED((float)lua_tonumber(L, 5)), toFIXED((float)lua_tonumber(L, 6))};
//		VECTOR p2 = {toFIXED((float)lua_tonumber(L, 7)), toFIXED((float)lua_tonumber(L, 8)), toFIXED((float)lua_tonumber(L, 9))};
//		VECTOR p3 = {toFIXED((float)lua_tonumber(L, 10)), toFIXED((float)lua_tonumber(L, 11)), toFIXED((float)lua_tonumber(L, 12))};
//		VECTOR ans;
//		slBezier(p0, p1, p2, p3, toFIXED((float)lua_tonumber(L, 13)), ans);
//		lua_pushnumber(L, (float)ans[X] / 65536.0);
//		lua_pushnumber(L, (float)ans[Y] / 65536.0);
//		lua_pushnumber(L, (float)ans[Z] / 65536.0);
//		return 3;
//	}
//	return 0;
//}

static int lua_slBezier(lua_State *L)
{
	if(lua_gettop(L)==5 && lua_istable(L, 1) && lua_istable(L, 2) && lua_istable(L, 3) && lua_istable(L, 4) /*&& lua_isnumber(L, 5)*/ ) {
		VECTOR p0, p1, p2, p3, ans;
		getVECTORfromTable(L, 1, p0);
		getVECTORfromTable(L, 2, p1);
		getVECTORfromTable(L, 3, p2);
		getVECTORfromTable(L, 4, p3);
		slBezier(p0, p1, p2, p3, toFIXED((float)lua_tonumber(L, 5)), ans);
		setVECTORtoTable(L, ans);
		return 1;
	}
	return 0;
}

//	Inputs
//		FIXED object1[XYZ], FIXED size1, FIXED object2[XYZ], FIXED size2
//	Outputs
//		FIXED

//static int lua_slBallCollision(lua_State *L)
//{
//	if(lua_gettop(L)==8 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) && lua_isnumber(L, 4) && lua_isnumber(L, 5) && lua_isnumber(L, 6) && lua_isnumber(L, 7) && lua_isnumber(L, 8) ) {
//		FIXED object1[XYZ] = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		FIXED object2[XYZ] = {toFIXED((float)lua_tonumber(L, 5)), toFIXED((float)lua_tonumber(L, 6)), toFIXED((float)lua_tonumber(L, 7))};
//		lua_pushnumber(L, (float)slBallCollision(object1, toFIXED((float)lua_tonumber(L, 4)), object2, toFIXED((float)lua_tonumber(L, 8))) / 65536.0);
//		return 1;
//	}
//	return 0;
//}

static int lua_slBallCollision(lua_State *L)
{
	if(lua_gettop(L)==4 && lua_istable(L, 1) /*&& lua_isnumber(L, 2)*/ && lua_istable(L, 3) /*&& lua_isnumber(L, 4)*/ ) {
		FIXED object1[XYZ], object2[XYZ];
		getVECTORfromTable(L, 1, object1);
		getVECTORfromTable(L, 3, object2);
		lua_pushnumber(L, fromFIXED(slBallCollision(object1, toFIXED((float)lua_tonumber(L, 2)), object2, toFIXED((float)lua_tonumber(L, 4))) ) );
		return 1;
	}
	return 0;
}

//	Inputs
//		FIXED object[XYZ], FIXED size
//	Outputs
//		FIXED

//static int lua_slCheckOnScreen(lua_State *L)
//{
//	if(lua_gettop(L)==4 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) && lua_isnumber(L, 4) ) {
//		FIXED object[XYZ] = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		FIXED ret = slCheckOnScreen(object, toFIXED((float)lua_tonumber(L, 4)));
//		lua_pushboolean(L, ret == -1);//Returns -1 (FFFFFFFF) when the object is in front of the screen
//		lua_pushboolean(L, ret == -2);//Returns -2 (FFFFFFFE) when the object is outside the upper-lower, left-right clipping boundaries
//		lua_pushboolean(L, (float)ret / 65536.0);//Returns the Z position when within the screen.
//		return 3;
//	}
//	return 0;
//}

static int lua_slCheckOnScreen(lua_State *L)
{
	if(lua_gettop(L)==2 && lua_istable(L, 1) /*&& lua_isnumber(L, 2)*/ ) {
		FIXED object[XYZ];
		getVECTORfromTable(L, 1, object);
		FIXED ret = slCheckOnScreen(object, toFIXED((float)lua_tonumber(L, 2)));
		lua_pushboolean(L, fromFIXED(ret) );//Returns the Z position when within the screen.
		lua_pushboolean(L, ret == -1);//Returns -1 (FFFFFFFF) when the object is in front of the screen
		lua_pushboolean(L, ret == -2);//Returns -2 (FFFFFFFE) when the object is outside the upper-lower, left-right clipping boundaries
		return 3;
	}
	return 0;
}

//	Inputs
//		FIXED size
//	Outputs
//		FIXED

static int lua_slCheckOnScreen0(lua_State *L)
{
	if(lua_gettop(L)==1 /*&& lua_isnumber(L, 1)*/ ) {
		FIXED ret = slCheckOnScreen0(toFIXED((float)lua_tonumber(L, 1)));
		lua_pushboolean(L, fromFIXED(ret) );//Returns the Z position when within the screen.
		lua_pushboolean(L, ret == -1);//Returns -1 (FFFFFFFF) when the object is in front of the screen
		lua_pushboolean(L, ret == -2);//Returns -2 (FFFFFFFE) when the object is outside the upper-lower, left-right clipping boundaries
		return 3;
	}
	return 0;
}

//void slSlaveFunc(void (*func)(), void *par)

//	Inputs
//		FIXED pos[XYZ]
//	Outputs
//		FIXED ret, Sint32 ans[XY]

//static int lua_slConvert3Dto2D(lua_State *L)
//{
//	if(lua_gettop(L)==3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) ) {
//		FIXED pos[XYZ] = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		Sint32 ans[XY];
//		lua_pushnumber(L, (float)slConvert3Dto2D(pos, ans) / 65536.0);
//		lua_pushinteger(L, ans[X]);
//		lua_pushinteger(L, ans[Y]);
//		return 3;
//	}
//	return 0;
//}

static int lua_slConvert3Dto2D(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_istable(L, 1) ) {
		FIXED pos[XYZ];
		Sint32 ans[XY];
		getVECTORfromTable(L, 1, pos);
		lua_pushnumber(L, fromFIXED(slConvert3Dto2D(pos, ans) ) );
		lua_createtable (L, 0, 2);
		setfield_Integer(L, ans[X], "X");
		setfield_Integer(L, ans[Y], "Y");
		return 2;
	}
	return 0;
}

//	Inputs
//		FIXED pos[XYZ]
//	Outputs
//		FIXED ret, FIXED ans[XY]

//static int lua_slConvert3Dto2DFX(lua_State *L)
//{
//	if(lua_gettop(L)==3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) ) {
//		FIXED pos[XYZ] = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		FIXED ans[XY];
//		lua_pushnumber(L, (float)slConvert3Dto2DFX(pos, ans) / 65536.0);
//		lua_pushnumber(L, (float)ans[X] / 65536.0);
//		lua_pushnumber(L, (float)ans[Y] / 65536.0);
//		return 3;
//	}
//	return 0;
//}

static int lua_slConvert3Dto2DFX(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_istable(L, 1) ) {
		FIXED pos[XYZ];
		FIXED ans[XY];
		getVECTORfromTable(L, 1, pos);
		lua_pushnumber(L, fromFIXED(slConvert3Dto2DFX(pos, ans) ) );
		lua_createtable (L, 0, 2);
		setfield_Number(L, fromFIXED(ans[X]), "X");
		setfield_Number(L, fromFIXED(ans[Y]), "Y");
		return 2;
	}
	return 0;
}

//	Inputs
//		FIXED , FIXED , FIXED
//	Outputs
//		FIXED ret, FIXED ans[XYZ]

//static int lua_slCalcPoint(lua_State *L)
//{
//	if(lua_gettop(L)==3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) ) {
//		FIXED ans[XYZ];
//		lua_pushnumber(L, (float)slCalcPoint(toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3)), ans) / 65536.0);
//		lua_pushnumber(L, (float)ans[X] / 65536.0);
//		lua_pushnumber(L, (float)ans[Y] / 65536.0);
//		lua_pushnumber(L, (float)ans[Z] / 65536.0);
//		return 4;
//	}
//	return 0;
//}



static int lua_slCalcPoint(lua_State *L)
{
	if(lua_gettop(L)==3 /*&& lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3)*/ ) {
		FIXED ans[XYZ];
		lua_pushnumber(L, fromFIXED(slCalcPoint(toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3)), ans) ) );
		setVECTORtoTable(L, ans);
		return 2;
	}
	return 0;
}

//	Inputs
//		VECTOR vct
//	Outputs
//		FIXED ret, VECTOR ans

//static int lua_slCalcVector(lua_State *L)
//{
//	if(lua_gettop(L)==3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) ) {
//		VECTOR vct = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		VECTOR ans;
//		lua_pushnumber(L, (float)slCalcVector(vct, ans) / 65536.0);
//		lua_pushnumber(L, (float)ans[X] / 65536.0);
//		lua_pushnumber(L, (float)ans[Y] / 65536.0);
//		lua_pushnumber(L, (float)ans[Z] / 65536.0);
//		return 4;
//	}
//	return 0;
//}

static int lua_slCalcVector(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_istable(L, 1) ) {
		VECTOR vct, ans;
		getVECTORfromTable(L, 1, vct);

		lua_pushnumber(L, fromFIXED(slCalcVector(vct, ans) ) );
		setVECTORtoTable(L, ans);
		return 2;
	}
	return 0;
}


//	Inputs
//		VECTOR vct
//	Outputs
//		none

//static int lua_slLight(lua_State *L)
//{
//	if(lua_gettop(L)==3 && lua_isnumber(L, 1) && lua_isnumber(L, 2) && lua_isnumber(L, 3) ) {
//		VECTOR vct = {toFIXED((float)lua_tonumber(L, 1)), toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3))};
//		slLight(vct);
//	}
//	return 0;
//}


static int lua_slLight(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_istable(L, 1) ) {
		VECTOR light;
		getVECTORfromTable(L, 1, light);
		slLight(light);
	}
	return 0;
}

//	Inputs
//		 XPDATA *, FIXED *
//	Outputs
//		none

//static int lua_slPutPolygonX(lua_State *L)
//{
//	if(lua_gettop(L)==4 && (lua_isuserdata(L, 1) || lua_isnumber(L, 1)) && lua_isnumber(L, 2) && lua_isnumber(L, 3) && lua_isnumber(L, 4) ) {
//		FIXED light[XYZ] = {toFIXED((float)lua_tonumber(L, 2)), toFIXED((float)lua_tonumber(L, 3)), toFIXED((float)lua_tonumber(L, 4))};
//		slPutPolygonX(lua_isuserdata(L,1) ? (XPDATA *)lua_touserdata(L,1): (XPDATA *)lua_tointeger(L,1), light);
//	}
//	return 0;
//}

static int lua_slPutPolygonX(lua_State *L)
{
	if(lua_gettop(L)==2 && lua_istable(L, 2) ) {
		FIXED light[XYZ];
		XPDATA *p = lua_isuserdata(L,1) ? (XPDATA *)lua_touserdata(L,1): (XPDATA *)lua_tointeger(L,1);
		if(p != NULL) {
			getVECTORfromTable(L, 2, light);		
			slPutPolygonX(p, light);
		}
	}
	return 0;
}


void lua_SGL_register(lua_State *L)
{
	if(L != NULL) {
		lua_register(L, "slPrint", lua_slPrint);
		lua_register(L, "slPrintHex", lua_slPrintHex);
		lua_register(L, "slPrintFX", lua_slPrintFX);
		lua_register(L, "slPCMOn", lua_slPCMOn);
		lua_register(L, "slPCMOff", lua_slPCMOff);
		lua_register(L, "slPCMStat", lua_slPCMStat);
		lua_register(L, "slInitSystem", lua_slInitSystem);		
		lua_register(L, "slGetTranslate", lua_slGetTranslate);
		lua_register(L, "slLookAt", lua_slLookAt);
		lua_register(L, "slCalcPoint", lua_slCalcPoint);
		lua_register(L, "slCalcVector", lua_slCalcVector);
		lua_register(L, "slInnerProduct", lua_slInnerProduct);
		lua_register(L, "slRandom", lua_slRandom);
		lua_register(L, "slBezier", lua_slBezier);
		lua_register(L, "slBallCollision", lua_slBallCollision);
		lua_register(L, "slCheckOnScreen", lua_slCheckOnScreen);
		lua_register(L, "slCheckOnScreen0", lua_slCheckOnScreen0);
		lua_register(L, "slConvert3Dto2D", lua_slConvert3Dto2D);
		lua_register(L, "slConvert3Dto2DFX", lua_slConvert3Dto2DFX);
		lua_register(L, "slDispHex", lua_slDispHex);
		lua_register(L, "slPrintMatrix", lua_slPrintMatrix);
		lua_register(L, "slExtendScroll", lua_slExtendScroll);
		lua_register(L, "sl16MapRA", lua_sl16MapRA);
		lua_register(L, "sl16MapRB", lua_sl16MapRB);
		lua_register(L, "slLight", lua_slLight);
		lua_register(L, "slPutPolygonX", lua_slPutPolygonX);
//
//		lua_register(L, "", lua_);
	}
}
