#include "lua.h"
#include "lauxlib.h"

//#include "def.h"
#include "errfunc.h"

#include <SEGA_GFS.H>
#include <SEGA_STM.H>


// taken from smpcm2.c

/* Xg[O[v̂hc */
static StmGrpHn grp_hd;


static Bool stmInit(Sint32 grpmax, Sint32 stmmax, void *stm_work)
{
	Bool ret = 0;// holds all the Bool return values of called functions
	/* Xg[VXȅ */
	ret |= STM_Init(grpmax, stmmax, stm_work);

	/* G[֐̐ݒ */
	STM_SetErrFunc(errStmFunc, NULL);

	/* Xg[O[ṽI[v */
	grp_hd = STM_OpenGrp();
	if(grp_hd != NULL) {
		ret |= 1 << 1;
		ret |= STM_SetLoop(grp_hd, STM_LOOP_DFL, STM_LOOP_ENDLESS) << 2;
		ret |= STM_SetExecGrp(grp_hd) << 3;
//		ret |= STM_NwSetExecGrp(grp_hd) << 3;// no change for now
	}
	return ret;
}

static StmHn stmOpen(char *fname)
{
    Sint32 fid;
	StmKey key;

    /* t@Ct@Cʎq߂ */
    fid = GFS_NameToId((Sint8 *)fname);
	STM_KEY_FN(&key) = STM_KEY_CN(&key) = STM_KEY_SMMSK(&key) = 
		STM_KEY_SMVAL(&key) = STM_KEY_CIMSK(&key) = STM_KEY_CIVAL(&key) =
		STM_KEY_NONE;
	return STM_OpenFid(grp_hd, fid, &key, STM_LOOP_NOREAD);
}

static void stmClose(StmHn fp)
{
	STM_Close(fp);
}



//	Inputs
//		Sint32 grpmax, Sint32 stmmax
//  Outputs
//		result: stm_work, integer value 16 on success, less on error

static int lua_stmInit(lua_State *L)
{
	Sint32 grpmax = 12, stmmax = 24; // same default values like in smppcm2.c
	Uint8 *stm_work;
	Bool ret = 0;
	// inputs
	if(lua_gettop(L) >= 1) {
		if(lua_isnumber(L, 1))
			grpmax = (int)lua_tointeger(L, 1);
		if(lua_gettop(L) >= 2) {
			if(lua_isnumber(L, 2))
				stmmax = (int)lua_tointeger(L, 2);
		}
	}
	// do work
	stm_work = (Uint8 *)lua_newuserdata(L, STM_WORK_SIZE(grpmax, stmmax));
	if(stm_work != NULL) {
		ret |= 1 << 4;
		ret |= stmInit(grpmax, stmmax, stm_work);
	}
	// outputs
	lua_pushinteger(L, ret);
	return 2;
}



//	Inputs
//		char *filename
//	Outputs
//		on success	:	StmHn
//		on error	:	none

static int lua_stmOpen(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_isstring(L, 1)) {
		StmHn stm = stmOpen((char *)lua_tostring(L, 1));
		if(stm != NULL) {
			lua_pushlightuserdata(L, stm);
			return 1;
		}
	}
	return 0;
}



//	Inputs
//		StmHn
//	Outputs
//		none

static int lua_stmClose(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1))
		stmClose((StmHn)lua_touserdata(L, 1));
	return 0;
}



//	Inputs
//		StmHn
//	Outputs
//		none

static int lua_STM_ResetTrBuf(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1))
		STM_ResetTrBuf((StmHn)lua_touserdata(L, 1));
	return 0;
}




//	Inputs
//		none (StmGrpHn, actually)
//	Outputs
//		Sint32 Stream Access status

static int lua_STM_GetStmNum(lua_State *L)
{
	lua_pushinteger(L, STM_GetStmNum(grp_hd));
	return 1;
}


//	Inputs
//		StmHn
//	Outputs
//		none

static int lua_STM_IsComplete(lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1)) {
		lua_pushboolean(L, STM_IsComplete((StmHn)lua_touserdata(L, 1)));
		return 1;
	}
	return 0;
}




void lua_STM_register(lua_State *L)
{
	if(L != NULL) {
		lua_register(L, "stmInit" 		, lua_stmInit);
		lua_register(L, "stmOpen" 		, lua_stmOpen);
		lua_register(L, "stmClose"		, lua_stmClose);
		lua_register(L, "STM_Close"		, lua_stmClose);
		lua_register(L, "STM_ResetTrBuf", lua_STM_ResetTrBuf);
		lua_register(L, "STM_GetStmNum"	, lua_STM_GetStmNum);
		lua_register(L, "STM_IsComplete", lua_STM_IsComplete);
	}
}

// example code for making it public as a lua library:
//static const struct luaL_reg mylib [] = {
//	{"dir", l_dir},
//	{NULL, NULL}  /* sentinel */
//};
//int luaopen_mylib (lua_State *L) {
//	luaL_openlib(L, "mylib", mylib, 0);//0 = not using upvalues
//	return 1;
//}


